import { Selector } from '@ngrx/store';

import { NgRxStoreSelect } from '../service';

/**
 * Slice state from the store.
 */
export function Select<State = any, Value = any>(selector: Selector<State, Value>): (target: any, name: string) => void;
export function Select<State = any, Value = any>(selectorOrFeature?: string, ...paths: string[]): (target: any, name: string) => void;
export function Select<State = any, Value = any>(selectorOrFeature?: string | Selector<State, Value>, ...paths: string[]) {
  return function(target: any, name: string): void {
    const selectorFnName = '__' + name + '__selector';
    let fn: Selector<State, Value>;
    // Nothing here? Use propery name as selector
    if (!selectorOrFeature) {
      selectorOrFeature = name;
    }

    // Handle string vs Selector<TState, TValue>
    if (typeof selectorOrFeature === 'string') {
      const propsArray = paths.length ? [selectorOrFeature, ...paths] : selectorOrFeature.split('.');
      fn = fastPropGetter(propsArray);
    } else {
      fn = selectorOrFeature;
    }

    const createSelect = () => {
      const store = NgRxStoreSelect.store;
      if (!store) {
        throw new Error('NgRxStoreSelect not connected to store!');
      }
      return store.select(fn);
    };

    if (target[selectorFnName]) {
      throw new Error('You cannot use @Select decorator and a ' + selectorFnName + ' property.');
    }

    // Redefine property
    if (delete target[name]) {
      Object.defineProperty(target, selectorFnName, {
        writable: true,
        enumerable: false,
        configurable: true
      });

      Object.defineProperty(target, name, {
        get: function() {
          return target[selectorFnName] || (target[selectorFnName] = createSelect.apply(target));
          // return this[selectorFnName] || (this[selectorFnName] = createSelect.apply(this));
        },
        enumerable: true,
        configurable: true
      });
    }
  };
}

/**
 * The generated function is faster than:
 * - pluck (Observable operator)
 * - memoize (old ngrx-actions implementation)
 * - MemoizedSelector (ngrx)
 */
export function fastPropGetter(paths: string[]): (x: any) => any {
  const segments = paths;
  let seg = 'store.' + segments[0],
    i = 0;
  const l = segments.length;
  let expr = seg;
  while (++i < l) {
    expr = expr + ' && ' + (seg = seg + '.' + segments[i]);
  }
  const fn = new Function('store', 'return ' + expr + ';');
  return <(x: any) => any>fn;
}
